from jabstract import jabstract

result_payload = jabstract({
    'command': '',
    'encoding': 'json',
    'result': {}
})

vlan_data = jabstract({
    'dynamic': False,
    'interfaces': {'Cpu': {'privatePromoted': False}},
    'name': 'VLAN0123',
    'status': 'active'
})

interface_address = jabstract({
    'broadcastAddress': '255.255.255.255',
    'dhcp': False,
    'primaryIp': {'address': '192.168.11.1', 'maskLen': 29},
    'secondaryIps': {
        '192.168.12.1': {'address': '192.168.12.1', 'maskLen': 29},
        '192.168.13.1': {'address': '192.168.13.1', 'maskLen': 29}
    },
    'secondaryIpsOrderedList': [
        {'address': '192.168.12.1', 'maskLen': 29},
        {'address': '192.168.13.1', 'maskLen': 29}
    ],
    'virtualIp': {'address': '0.0.0.0', 'maskLen': 0},
    'virtualSecondaryIps': {},
    'virtualSecondaryIpsOrderedList': []
})

interface_vlan_data = jabstract({
    'bandwidth': 0,
    'burnedInAddress': 'fa:16:3e:02:b9:e2',
    'description': '',
    'forwardingModel': 'routed',
    'hardware': 'vlan',
    'interfaceAddress': [interface_address()],
    'interfaceStatus': 'notconnect',
    'lastStatusChangeTimestamp': 1539974209.994736,
    'lineProtocolStatus': 'lowerLayerDown',
    'mtu': 1500,
    'name': 'Vlan123',
    'physicalAddress': 'fa:16:3e:02:b9:e2'
})

interface_data = jabstract({
    'autoNegotiate': 'off',
    'bandwidth': 0,
    'burnedInAddress': 'fa:16:3e:0e:ee:81',
    'description': '',
    'duplex': 'duplexFull',
    'forwardingModel': 'bridged',
    'hardware': 'ethernet',
    'interfaceAddress': [],
    'interfaceCounters': {
        'counterRefreshTime': 1541082601.226264,
        'inBroadcastPkts': 0,
        'inDiscards': 0,
        'inMulticastPkts': 30902,
        'inOctets': 2183607,
        'inTotalPkts': 30948,
        'inUcastPkts': 46,
        'inputErrorsDetail': {
            'alignmentErrors': 0,
            'fcsErrors': 0,
            'giantFrames': 0,
            'runtFrames': 0,
            'rxPause': 0,
            'symbolErrors': 0},
        'linkStatusChanges': 4,
        'outBroadcastPkts': 0,
        'outDiscards': 0,
        'outMulticastPkts': 447388,
        'outOctets': 32423850,
        'outUcastPkts': 0,
        'outputErrorsDetail': {
            'collisions': 0,
            'deferredTransmissions': 0,
            'lateCollisions': 0,
            'txPause': 0},
        'totalInErrors': 0,
        'totalOutErrors': 0},
    'interfaceStatistics': {
        'inBitsRate': 0.0,
        'inPktsRate': 0.0,
        'outBitsRate': 0.0,
        'outPktsRate': 0.0,
        'updateInterval': 300.0},
    'interfaceStatus': 'disabled',
    'lastStatusChangeTimestamp': 1541080271.3714354,
    'lineProtocolStatus': 'down',
    'loopbackMode': 'loopbackNone',
    'mtu': 9214,
    'name': 'Ethernet1',
    'physicalAddress': 'fa:16:3e:0e:ee:81'
})

switchport_data = jabstract({
    'accessVlanId': 1,
    'accessVlanName': 'default',
    'dot1qVlanTagRequired': False,
    'dot1qVlanTagRequiredStatus': False,
    'dynamicAllowedVlans': {},
    'dynamicTrunkGroups': [],
    'macLearning': True,
    'mode': 'access',
    'sourceportFilterMode': 'enabled',
    'staticTrunkGroups': [],
    'tpid': '0x8100',
    'tpidStatus': True,
    'trunkAllowedVlans': '800,805',
    'trunkingNativeVlanId': 1,
    'trunkingNativeVlanName': 'default'
})


def show_interfaces(*interface_data):
    return {'interfaces': {i["name"]: i for i in interface_data}}



# Copyright 2019 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from hamcrest import assert_that, equal_to

from netman.core.objects.exceptions import UnknownVlan
from tests.adapters.compliance_test_case import ComplianceTestCase


class RemoveVlanTest(ComplianceTestCase):
    _dev_sample = "brocade"

    def setUp(self):
        super(RemoveVlanTest, self).setUp()
        self.client.add_vlan(1000)

    def tearDown(self):
        self.janitor.remove_vlan(1000)
        super(RemoveVlanTest, self).tearDown()

    def test_removes_vlan_from_get_vlan(self):
        self.client.remove_vlan(1000)

        with self.assertRaises(UnknownVlan):
            self.client.get_vlan(1000)

    def test_removes_vlan_raise_when_vlan_is_already_removed(self):
        self.client.remove_vlan(1000)

        with self.assertRaises(UnknownVlan):
            self.client.remove_vlan(1000)

    def test_removes_vlan_is_removed_from_list(self):
        vlan_count = len(self.client.get_vlans())
        self.client.remove_vlan(1000)

        assert_that(len(self.client.get_vlans()), equal_to(vlan_count - 1))



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from contextlib import contextmanager
from functools import wraps

from netman.core.objects.backward_compatible_switch_operations import BackwardCompatibleSwitchOperations


def not_implemented(func):
    @wraps(func)
    def func_wrapper(*args, **kwargs):
        raise NotImplementedError("'{}' is not implemented".format(func.func_name))
    return func_wrapper


class SwitchOperations(BackwardCompatibleSwitchOperations):

    @not_implemented
    def connect(self):
        pass

    @not_implemented
    def disconnect(self):
        pass

    @not_implemented
    def transaction(self):
        pass

    @not_implemented
    def start_transaction(self):
        pass

    @not_implemented
    def commit_transaction(self):
        pass

    @not_implemented
    def rollback_transaction(self):
        pass

    @not_implemented
    def end_transaction(self):
        pass

    @not_implemented
    def get_vlan(self, number):
        pass

    @not_implemented
    def get_vlans(self):
        pass

    @not_implemented
    def add_vlan(self, number, name=None):
        pass

    @not_implemented
    def remove_vlan(self, number):
        pass

    @not_implemented
    def get_interface(self, interface_id):
        pass

    @not_implemented
    def get_interfaces(self):
        pass

    @not_implemented
    def set_access_vlan(self, interface_id, vlan):
        pass

    @not_implemented
    def unset_interface_access_vlan(self, interface_id):
        pass

    @not_implemented
    def set_access_mode(self, interface_id):
        pass

    @not_implemented
    def set_trunk_mode(self, interface_id):
        pass

    @not_implemented
    def add_trunk_vlan(self, interface_id, vlan):
        pass

    @not_implemented
    def remove_trunk_vlan(self, interface_id, vlan):
        pass

    @not_implemented
    def set_interface_state(self, interface_id, state):
        pass

    @not_implemented
    def unset_interface_state(self, interface_id):
        pass

    @not_implemented
    def set_interface_auto_negotiation_state(self, interface_id, negotiation_state):
        pass

    @not_implemented
    def unset_interface_auto_negotiation_state(self, interface_id):
        pass

    @not_implemented
    def set_interface_native_vlan(self, interface_id, vlan):
        pass

    @not_implemented
    def unset_interface_native_vlan(self, interface_id):
        pass

    @not_implemented
    def reset_interface(self, interface_id):
        pass

    @not_implemented
    def get_vlan_interfaces(self, vlan_number):
        pass

    @not_implemented
    def add_ip_to_vlan(self, vlan_number, ip_network):
        pass

    @not_implemented
    def remove_ip_from_vlan(self, vlan_number, ip_network):
        pass

    @not_implemented
    def set_vlan_access_group(self, vlan_number, direction, name):
        pass

    @not_implemented
    def unset_vlan_access_group(self, vlan_number, direction):
        pass

    @not_implemented
    def set_vlan_vrf(self, vlan_number, vrf_name):
        pass

    @not_implemented
    def unset_vlan_vrf(self, vlan_number):
        pass

    @not_implemented
    def set_interface_description(self, interface_id, description):
        pass

    @not_implemented
    def unset_interface_description(self, interface_id):
        pass

    @not_implemented
    def edit_interface_spanning_tree(self, interface_id, edge=None):
        pass

    @not_implemented
    def add_bond(self, number):
        pass

    @not_implemented
    def remove_bond(self, number):
        pass

    @not_implemented
    def get_bond(self, number):
        pass

    @not_implemented
    def get_bonds(self):
        pass

    @not_implemented
    def add_interface_to_bond(self, interface, bond_number):
        pass

    @not_implemented
    def remove_interface_from_bond(self, interface):
        pass

    @not_implemented
    def set_bond_link_speed(self, number, speed):
        pass

    @not_implemented
    def set_bond_description(self, number, description):
        pass

    @not_implemented
    def unset_bond_description(self, number):
        pass

    @not_implemented
    def set_bond_trunk_mode(self, number):
        pass

    @not_implemented
    def set_bond_access_mode(self, number):
        pass

    @not_implemented
    def add_bond_trunk_vlan(self, number, vlan):
        pass

    @not_implemented
    def remove_bond_trunk_vlan(self, number, vlan):
        pass

    @not_implemented
    def set_bond_native_vlan(self, number, vlan):
        pass

    @not_implemented
    def unset_bond_native_vlan(self, number):
        pass

    @not_implemented
    def edit_bond_spanning_tree(self, number, edge=None):
        pass

    @not_implemented
    def add_vrrp_group(self, vlan_number, group_id, ips=None, priority=None, hello_interval=None, dead_interval=None,
                       track_id=None, track_decrement=None):
        pass

    @not_implemented
    def remove_vrrp_group(self, vlan_id, group_id):
        pass

    @not_implemented
    def add_dhcp_relay_server(self, vlan_number, ip_address):
        pass

    @not_implemented
    def remove_dhcp_relay_server(self, vlan_number, ip_address):
        pass

    @not_implemented
    def set_interface_lldp_state(self, interface_id, enabled):
        pass

    @not_implemented
    def set_vlan_arp_routing_state(self, vlan_number, state):
        pass

    @not_implemented
    def set_vlan_icmp_redirects_state(self, vlan_number, state):
        pass

    @not_implemented
    def set_vlan_unicast_rpf_mode(self, vlan_number, mode):
        pass

    @not_implemented
    def unset_vlan_unicast_rpf_mode(self, vlan_number):
        pass

    @not_implemented
    def get_versions(self):
        pass

    @not_implemented
    def set_interface_mtu(self, interface_id, size):
        pass

    @not_implemented
    def unset_interface_mtu(self, interface_id):
        pass

    @not_implemented
    def set_bond_mtu(self, number, size):
        pass

    @not_implemented
    def unset_bond_mtu(self, number):
        pass

    @not_implemented
    def set_vlan_ntp_state(self, vlan_number, state):
        pass

    @not_implemented
    def add_vlan_varp_ip(self, vlan_number, ip_network):
        pass

    @not_implemented
    def remove_vlan_varp_ip(self, vlan_number, ip_network):
        pass

    @not_implemented
    def set_vlan_load_interval(self, vlan_number, time_interval):
        pass

    @not_implemented
    def unset_vlan_load_interval(self, vlan_number):
        pass

    @not_implemented
    def set_vlan_mpls_ip_state(self, vlan_number, state):
        pass


class SwitchBase(SwitchOperations):
    def __init__(self, switch_descriptor):
        self.switch_descriptor = switch_descriptor
        self.logger = logging.getLogger("{module}.{hostname}".format(module=self.__module__, hostname=self.switch_descriptor.hostname))
        self.connected = False
        self.in_transaction = False

    def connect(self):
        self._connect()
        self.connected = True

    def disconnect(self):
        self._disconnect()
        self.connected = False

    def start_transaction(self):
        self._start_transaction()
        self.in_transaction = True

    def end_transaction(self):
        self._end_transaction()
        self.in_transaction = False

    def _connect(self):
        """
        Adpapters should implement this rather than connect
        """
        raise NotImplementedError()

    def _disconnect(self):
        """
        Adpapters should implement this rather than disconnect
        """
        raise NotImplementedError()

    def _start_transaction(self):
        """
        Adpapters should implement this rather than connect
        """
        raise NotImplementedError()

    def _end_transaction(self):
        """
        Adpapters should implement this rather than disconnect
        """
        raise NotImplementedError()

    @contextmanager
    def transaction(self):
        self.start_transaction()
        try:
            yield self
            self.commit_transaction()
        except Exception as e:
            if self.logger:
                self.logger.exception(e)
            self.rollback_transaction()
            raise
        finally:
            self.end_transaction()



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from ncclient.xml_ import to_ele, new_ele

from netman.adapters.switches.juniper.base import interface_speed, interface_replace, interface_speed_update, \
    first_text, bond_name, Juniper, first, value_of, parse_range, to_range
from netman.core.objects.exceptions import BadVlanName, BadVlanNumber, VlanAlreadyExist, UnknownVlan


def netconf(switch_descriptor, *args, **kwargs):
    return Juniper(switch_descriptor, custom_strategies=JuniperCustomStrategies(), *args, **kwargs)


class JuniperCustomStrategies(object):
    def get_interface_port_mode_update_element(self, mode):
        return to_ele("<port-mode>{}</port-mode>".format(mode))

    def get_vlan_member_update_element(self, vlan):
        return to_ele("<members>{}</members>".format(vlan))

    def get_port_mode_node_in_inteface_node(self, interface_node):
        return interface_node.xpath("unit/family/ethernet-switching/port-mode")

    def add_enslave_to_bond_operations(self, update, interface, bond):
        ether_options = [
            to_ele("""
                <ieee-802.3ad>
                    <bundle>{0}</bundle>
                </ieee-802.3ad>
            """.format(bond_name(bond.number)))]

        if bond.link_speed is not None:
            ether_options.append(interface_speed(bond.link_speed))

        update.add_interface(interface_replace(interface, *ether_options))

    def add_update_bond_members_speed_operations(self, update, slave_nodes, speed):
        for interface_node in slave_nodes:
            update.add_interface(interface_speed_update(first_text(interface_node.xpath("name")), speed))

    def add_update_vlans(self, update, number, name):
        update.add_vlan(self.vlan_update(number, name), "vlans")

    def remove_update_vlans(self, update, vlan_name):
        update.add_vlan(self.vlan_removal(vlan_name), "vlans")

    def get_interface_trunk_native_vlan_id_node(self, interface):
        return interface.xpath("unit/family/ethernet-switching/native-vlan-id")

    def set_native_vlan_id_node(self, interface_node, native_vlan_id_node):
        return interface_node.xpath("//ethernet-switching")[0].append(native_vlan_id_node)

    def get_protocols_interface_name(self, interface_name):
        return "{}.0".format(interface_name)

    def all_vlans(self):
        return new_ele("vlans")

    def one_vlan_by_vlan_id(self, vlan_id):
        def m():
            return to_ele("""
            <vlans>
                <vlan>
                    <vlan-id>{}</vlan-id>
                </vlan>
            </vlans>
        """.format(vlan_id))

        return m

    def vlan_update(self, number, description):
        content = to_ele("""
            <vlan>
                <name>VLAN{0}</name>
                <vlan-id>{0}</vlan-id>
            </vlan>
        """.format(number))

        if description is not None:
            content.append(to_ele("<description>{}</description>".format(description)))
        return content

    def vlan_removal(self, name):
        return to_ele("""
        <vlan operation="delete">
            <name>{}</name>
        </vlan>""".format(name))

    def vlan_nodes(self, config):
        return config.xpath("data/configuration/vlans/vlan")

    def vlan_node(self, config, number):
        vlan_node = first(config.xpath("data/configuration/vlans/vlan/vlan-id[text()=\"{}\"]/..".format(number)))

        if vlan_node is None:
            raise UnknownVlan(number)
        return vlan_node

    def manage_update_vlan_exception(self, message, number):
        if "being used by" in message:
            raise VlanAlreadyExist(number)
        elif "not within range" in message:
            if message.startswith("Value"):
                raise BadVlanNumber()
            elif message.startswith("Length"):
                raise BadVlanName()

    def interface_update(self, name, unit, attributes=None, vlan_members=None):
        content = to_ele("""
            <interface>
                <name>{interface}</name>
                <unit>
                    <name>{unit}</name>
                    <family>
                        <ethernet-switching>
                        </ethernet-switching>
                    </family>
                </unit>
            </interface>
            """.format(interface=name, unit=unit))
        ethernet_switching_node = first(content.xpath("//ethernet-switching"))

        for attribute in (attributes if attributes is not None else []):
            ethernet_switching_node.append(attribute)

        if vlan_members:
            vlan = new_ele("vlan")
            for attribute in vlan_members:
                vlan.append(attribute)
            ethernet_switching_node.append(vlan)

        return content

    def get_l3_interface(self, vlan_node):
        if_name_node = first(vlan_node.xpath("l3-interface"))
        if if_name_node is not None:
            return if_name_node.text.split(".")
        else:
            return None, None

    def list_vlan_members(self, interface_node, config):
        vlans = set()
        for members in interface_node.xpath("unit/family/ethernet-switching/vlan/members"):
            vlan_id = value_of(config.xpath('data/configuration/vlans/vlan/name[text()="{}"]/../vlan-id'.format(members.text)), transformer=int)
            if vlan_id:
                vlans = vlans.union([vlan_id])
            else:
                vlans = vlans.union(parse_range(members.text))
        return sorted(vlans)

    def update_vlan_members(self, interface_node, vlan_members, vlan):
        if interface_node is not None:
            for members in interface_node.xpath("unit/family/ethernet-switching/vlan/members"):
                vlan_members.append(to_ele('<members operation="delete">{}</members>'.format(members.text)))
        vlan_members.append(to_ele("<members>{}</members>".format(vlan)))

    def craft_members_modification_to_remove_vlan(self, interface_node, vlan_name, number):
        members_modifications = []
        for vlan_members_node in interface_node.xpath("unit/family/ethernet-switching/vlan/members"):
            if vlan_members_node.text == vlan_name:
                members_modifications.append(to_ele("<members operation=\"delete\">{}</members>".format(vlan_members_node.text)))
            else:
                vlan_list = parse_range(vlan_members_node.text)
                if number in vlan_list:
                    members_modifications.append(to_ele("<members operation=\"delete\">{}</members>".format(vlan_members_node.text)))

                    below = vlan_list[:vlan_list.index(number)]
                    if len(below) > 0:
                        members_modifications.append(to_ele("<members>{}</members>".format(to_range(below))))

                    above = vlan_list[vlan_list.index(number) + 1:]
                    if len(above) > 0:
                        members_modifications.append(to_ele("<members>{}</members>".format(to_range(above))))

        return members_modifications

    def interface_vlan_members_update(self, name, unit, members_modification):
        content = to_ele("""
            <interface>
                <name>{}</name>
                <unit>
                    <name>{}</name>
                    <family>
                        <ethernet-switching>
                            <vlan />
                        </ethernet-switching>
                    </family>
                </unit>
            </interface>
            """.format(name, unit))

        vlan_node = first(content.xpath("//vlan"))
        for m in members_modification:
            vlan_node.append(m)

        return content

    def parse_vrrp_groups(self, interface_unit):
        return []

    def parse_icmp_redirects(self, interface_unit):
        return []

    def get_delete_trunk_vlan_element(self):
        return self.get_delete_vlan_element()

    def get_delete_vlan_element(self):
        return to_ele('<vlan operation="delete" />')



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from flask import request

from netman.api.api_utils import BadRequest
from netman.core.objects.switch_descriptor import SwitchDescriptor


class SwitchApiBase(object):
    def __init__(self, switch_factory, sessions_manager):
        self.switch_factory = switch_factory
        self.sessions_manager = sessions_manager

    def hook_to(self, server):
        raise NotImplemented()

    @property
    def logger(self):
        return logging.getLogger(__name__)

    def _get_switch_descriptor_from_request_headers(self, hostname):
        headers_present = [h in request.headers for h in ['Netman-Model', 'Netman-Username', 'Netman-Password']]

        if not any(headers_present):
            return None

        if not all(headers_present):
            raise BadRequest('For anonymous switch usage, please specify headers: Netman-Model, Netman-Username and Netman-Password.')

        port = None
        if "Netman-Port" in request.headers:
            try:
                port = int(request.headers["Netman-Port"])
            except ValueError:
                raise BadRequest('Netman-Port optional header should be an integer')

        netman_server = request.headers.get("Netman-Proxy-Server", None)
        if netman_server is not None and "," in netman_server:
            netman_server = [e.strip() for e in netman_server.split(",")]

        self.logger.info("Anonymous Switch Access ({}) {}@{}".format(request.headers['Netman-Model'], request.headers['Netman-Username'], hostname))
        return SwitchDescriptor(
            hostname=hostname,
            model=request.headers['Netman-Model'],
            username=request.headers['Netman-Username'],
            password=request.headers['Netman-Password'],
            port=port,
            netman_server=netman_server
        )

    def resolve_switch(self, hostname):
        switch_descriptor = self._get_switch_descriptor_from_request_headers(hostname)

        if switch_descriptor:
            return self.switch_factory.get_switch_by_descriptor(switch_descriptor)
        return self.switch_factory.get_switch(hostname)

    def resolve_session(self, session_id):
        return self.sessions_manager.get_switch_for_session(session_id)



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import TestCase
import time

from flexmock import flexmock
from hamcrest import assert_that, is_
from mock import Mock
from netman.core.objects.exceptions import UnknownResource, \
    NetmanException, SessionAlreadyExists
from netman.core.objects.switch_descriptor import SwitchDescriptor
from netman.core.switch_sessions import SwitchSessionManager


class SwitchSessionManagerTest(TestCase):
    def setUp(self):
        self.switch_mock = flexmock()
        self.switch_mock.switch_descriptor = SwitchDescriptor('dell', "a_host")
        self.session_manager = SwitchSessionManager()

    def tearDown(self):
        for timer in self.session_manager.timers.values():
            timer.cancel()

    def test_open_session_generates_with_passed_session_id(self):
        self.session_manager.session_storage = flexmock()
        self.session_manager.session_storage.should_receive('add').with_args(
            'patate', self.switch_mock.switch_descriptor
        ).once()
        self.switch_mock.should_receive('connect').once()

        assert_that(self.session_manager.open_session(self.switch_mock, 'patate'), is_('patate'))
        assert_that(self.session_manager.get_switch_for_session('patate'), is_(self.switch_mock))

    def test_open_session_with_session_that_already_exists_raises_an_exception(self):
        self.session_manager.session_storage = flexmock()
        self.session_manager.session_storage.should_receive('add').with_args(
            'patate', self.switch_mock.switch_descriptor
        ).once()
        self.switch_mock.should_receive('connect').once()

        self.session_manager.open_session(self.switch_mock, 'patate')

        with self.assertRaises(SessionAlreadyExists):
            self.session_manager.open_session(self.switch_mock, 'patate')

    def test_close_session(self):
        self.session_manager.session_storage = flexmock()
        self.session_manager.session_storage.should_receive('add').with_args(
            'patate', self.switch_mock.switch_descriptor
        ).once()
        self.switch_mock.should_receive('connect').once()

        assert_that(self.session_manager.open_session(self.switch_mock, 'patate'), is_('patate'))
        assert_that(self.session_manager.get_switch_for_session('patate'), is_(self.switch_mock))

        self.switch_mock.should_receive('disconnect').once()
        self.session_manager.session_storage.should_receive('remove').with_args('patate')

        self.session_manager.close_session('patate')

        with self.assertRaises(UnknownResource):
            self.session_manager.get_switch_for_session('patate')

    def test_close_session_catches_exception_if_remote_remove_fails(self):
        self.session_manager.session_storage = flexmock()

        self.session_manager.session_storage.should_receive('add').with_args(
            'patate', self.switch_mock.switch_descriptor
        ).once().ordered()
        self.switch_mock.should_receive('connect').once()

        self.session_manager.open_session(self.switch_mock, 'patate')

        self.switch_mock.should_receive('disconnect').once()
        self.session_manager.session_storage.should_receive('remove').with_args('patate').and_raise(NetmanException)

        self.session_manager.close_session('patate')

    def test_session_should_close_itself_after_timeout(self):
        self.session_manager.session_inactivity_timeout = 0.01

        switch_mock = Mock()

        assert_that(self.session_manager.open_session(switch_mock, 'patate'), is_('patate'))
        assert_that(self.session_manager.get_switch_for_session('patate'), is_(switch_mock))

        time.sleep(0.02)

        assert_that(switch_mock.rollback_transaction.called, is_(False))

        with self.assertRaises(UnknownResource):
            self.session_manager.get_switch_for_session('patate')

    def test_session_timeout_should_reset_on_activity(self):
        self.session_manager.session_inactivity_timeout = 0.1

        switch_mock = Mock()

        assert_that(self.session_manager.open_session(switch_mock, 'patate'), is_('patate'))
        assert_that(self.session_manager.get_switch_for_session('patate'), is_(switch_mock))

        time.sleep(0.05)

        self.session_manager.keep_alive('patate')

        time.sleep(0.05)

        assert_that(switch_mock.rollback_transaction.called, is_(False))

        self.session_manager.keep_alive('patate')

        time.sleep(0.11)

        assert_that(switch_mock.rollback_transaction.called, is_(False))

        with self.assertRaises(UnknownResource):
            self.session_manager.get_switch_for_session('patate')

    def test_commit_transaction(self):
        self.session_manager.keep_alive = Mock()
        self.session_manager.session_storage = flexmock()

        self.session_manager.session_storage.should_receive('add').with_args(
            'patate', self.switch_mock.switch_descriptor
        ).once()
        self.switch_mock.should_receive('connect').once().ordered()

        session_id = self.session_manager.open_session(self.switch_mock, 'patate')

        self.switch_mock.should_receive('commit_transaction').once().ordered()

        self.assertEquals(session_id, 'patate')

        self.session_manager.commit_session(session_id)

        self.session_manager.keep_alive.assert_called_with(session_id)

    def test_rollback_session(self):
        self.session_manager.keep_alive = Mock()
        self.session_manager.session_storage = flexmock()

        self.session_manager.session_storage.should_receive('add').with_args(
            'patate', self.switch_mock.switch_descriptor
        ).once()
        self.switch_mock.should_receive('connect').once()

        session_id = self.session_manager.open_session(self.switch_mock, 'patate')

        self.assertEquals(session_id, 'patate')

        self.switch_mock.should_receive('rollback_transaction').once().ordered()

        self.session_manager.rollback_session(session_id)

        self.session_manager.keep_alive.assert_called_with(session_id)

    def test_unknown_session(self):
        with self.assertRaises(UnknownResource):
            self.session_manager.get_switch_for_session('patate')

    def test_add_session_catches_exception_if_remote_add_fails(self):
        self.session_manager.session_storage = flexmock()
        self.session_manager.session_storage.should_receive('add').with_args(
            'patate', self.switch_mock.switch_descriptor
        ).and_raise(NetmanException)
        self.switch_mock.should_receive('connect').once().ordered()

        self.session_manager.open_session(self.switch_mock, 'patate')

    def test_start_transaction(self):
        self.session_manager.keep_alive = Mock()
        self.session_manager.session_storage = flexmock()

        self.session_manager.session_storage.should_receive('add').with_args(
            'patate', self.switch_mock.switch_descriptor
        ).once()
        self.switch_mock.should_receive('connect').once().ordered()

        session_id = self.session_manager.open_session(self.switch_mock, 'patate')

        self.switch_mock.should_receive('start_transaction').once().ordered()

        self.assertEquals(session_id, 'patate')

        self.session_manager.start_transaction(session_id)

        self.session_manager.keep_alive.assert_called_with(session_id)

    def test_end_transaction(self):
        self.session_manager.keep_alive = Mock()
        self.session_manager.session_storage = flexmock()

        self.session_manager.session_storage.should_receive('add').with_args(
            'patate', self.switch_mock.switch_descriptor
        ).once()
        self.switch_mock.should_receive('connect').once().ordered()

        session_id = self.session_manager.open_session(self.switch_mock, 'patate')

        self.switch_mock.should_receive('end_transaction').once().ordered()

        self.assertEquals(session_id, 'patate')

        self.session_manager.end_transaction(session_id)

        self.session_manager.keep_alive.assert_called_with(session_id)



# Copyright 2015 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import __builtin__
import importlib
import json
import uuid
import warnings

import requests

from netman import raw_or_json
from netman.api import NETMAN_API_VERSION
from netman.api.objects import bond
from netman.api.objects import interface
from netman.api.objects import vlan
from netman.core.objects.access_groups import IN, OUT
from netman.core.objects.exceptions import NetmanException, UnknownSession
from netman.core.objects.interface_states import OFF, ON
from netman.core.objects.switch_base import SwitchBase


def factory(switch_descriptor):
    warnings.warn("Use SwitchFactory.get_switch_by_descriptor directly to instanciate a switch", DeprecationWarning)
    return RemoteSwitch(switch_descriptor)


class RemoteSwitch(SwitchBase):
    max_version = NETMAN_API_VERSION

    def __init__(self, switch_descriptor):
        super(RemoteSwitch, self).__init__(switch_descriptor)
        self.requests = requests
        self.session_id = None

        if isinstance(self.switch_descriptor.netman_server, list):
            self._proxy = self.switch_descriptor.netman_server[0]
            self._next_proxies = self.switch_descriptor.netman_server[1:]
        else:
            self._proxy = self.switch_descriptor.netman_server
            self._next_proxies = []

    def _connect(self):
        self.session_id = str(uuid.uuid4())
        self.logger.info("Requesting session {}".format(self.session_id))
        url = "{netman}/switches-sessions/{session_id}".format(netman=self._proxy, session_id=self.session_id)
        details = self.request()
        details['headers']['Netman-Session-Id'] = self.session_id
        self.validated(self.requests.post(
            url=url,
            data=json.dumps({'hostname': self.switch_descriptor.hostname}),
            headers=details['headers'])
        )
        self.logger.info("Obtained session {}".format(self.session_id))

    def _disconnect(self):
        self.logger.info("Ending session {}".format(self.session_id))
        url = "{netman}/switches-sessions/{session_id}".format(netman=self._proxy, session_id=self.session_id)
        session_id = self.session_id
        self.session_id = None
        self.validated(self.requests.delete(url=url, headers={'Netman-Verbose-Errors': "yes",
                                                              'Netman-Max-Version': str(self.max_version),
                                                              'Netman-Session-Id': session_id}))
        self.logger.info("Ended session {}".format(self.session_id))

    def _start_transaction(self):
        self.logger.info("Starting Transaction for session_id: {}".format(self.session_id))
        url = "{netman}/switches-sessions/{session_id}/actions".format(netman=self._proxy, session_id=self.session_id)
        self.validated(self.requests.post(url=url, headers={'Netman-Verbose-Errors': "yes",
                                                            'Netman-Max-Version': str(self.max_version),
                                                            'Netman-Session-Id': self.session_id}, data='start_transaction'))
        self.logger.info("Started Transaction for session_id: {}".format(self.session_id))

    def commit_transaction(self):
        self.logger.info("Commiting {}".format(self.session_id))
        url = "{netman}/switches-sessions/{session_id}/actions".format(netman=self._proxy, session_id=self.session_id)
        self.validated(self.requests.post(url=url, headers={'Netman-Verbose-Errors': "yes",
                                                            'Netman-Max-Version': str(self.max_version),
                                                            'Netman-Session-Id': self.session_id}, data='commit'))
        self.logger.info("Commited {}".format(self.session_id))

    def rollback_transaction(self):
        self.logger.info("Rollbacking {}".format(self.session_id))
        url = "{netman}/switches-sessions/{session_id}/actions".format(netman=self._proxy, session_id=self.session_id)
        self.validated(self.requests.post(url=url, headers={'Netman-Verbose-Errors': "yes",
                                                            'Netman-Max-Version': str(self.max_version),
                                                            'Netman-Session-Id': self.session_id}, data='rollback'))
        self.logger.info("Rollbacked {}".format(self.session_id))

    def _end_transaction(self):
        self.logger.info("Ending Transaction for session_id: {}".format(self.session_id))
        url = "{netman}/switches-sessions/{session_id}/actions".format(netman=self._proxy, session_id=self.session_id)
        self.validated(self.requests.post(url=url, headers={'Netman-Verbose-Errors': "yes",
                                                            'Netman-Max-Version': str(self.max_version),
                                                            'Netman-Session-Id': self.session_id}, data='end_transaction'))
        self.logger.info("Transaction ended for session_id: {}".format(self.session_id))

    def get_vlan(self, number):
        return vlan.to_core(self.get("/vlans/{}".format(number)).json())

    def get_vlans(self):
        return [vlan.to_core(row) for row in self.get("/vlans").json()]

    def get_interface(self, interface_id):
        return interface.to_core(self.get("/interfaces/{}".format(interface_id)).json())

    def get_interfaces(self):
        return [interface.to_core(row) for row in self.get("/interfaces").json()]

    def get_bond(self, number):
        reply = self.get('/bonds/{}'.format(number))
        return bond.to_core(reply.json(), version=reply.headers.get('Netman-Version'))

    def get_bonds(self):
        reply = self.get("/bonds")
        return [bond.to_core(row, version=reply.headers.get('Netman-Version')) for row in reply.json()]

    def add_vlan(self, number, name=None):
        data = {'number': number}
        if name is not None:
            data['name'] = name
        self.post("/vlans", data=data)

    def remove_vlan(self, number):
        self.delete("/vlans/{0}".format(str(number)))

    def get_vlan_interfaces(self, vlan_number):
        return self.get("/vlans/{}/interfaces".format(vlan_number)).json()

    def set_vlan_access_group(self, vlan_number, direction, name):
        self.put('/vlans/{vlan_number}/access-groups/{direction}'.format(
            vlan_number=vlan_number,
            direction={IN: 'in', OUT: 'out'}[direction]
        ), raw_data=name)

    def unset_vlan_access_group(self, vlan_number, direction):
        self.delete('/vlans/{vlan_number}/access-groups/{direction}'.format(
            vlan_number=vlan_number,
            direction={IN: 'in', OUT: 'out'}[direction]
        ))

    def add_ip_to_vlan(self, vlan_number, ip_network):
        self.post('/vlans/{vlan_number}/ips'.format(
            vlan_number=vlan_number
        ), raw_data=str(ip_network))

    def remove_ip_from_vlan(self, vlan_number, ip_network):
        self.delete('/vlans/{vlan_number}/ips/{ip_network}'.format(
            vlan_number=vlan_number,
            ip_network=ip_network
        ))

    def set_vlan_vrf(self, vlan_number, vrf_name):
        self.put('/vlans/{vlan_number}/vrf-forwarding'.format(
            vlan_number=vlan_number
        ), raw_data=str(vrf_name))

    def unset_vlan_vrf(self, vlan_number):
        self.delete('/vlans/{vlan_number}/vrf-forwarding'.format(vlan_number=vlan_number))

    def set_vlan_load_interval(self, vlan_number, time_interval):
        self.put('/vlans/{vlan_number}/load-interval'.format(
            vlan_number=vlan_number
        ), raw_data=str(time_interval))

    def unset_vlan_load_interval(self, vlan_number):
        self.delete('/vlans/{vlan_number}/load-interval'.format(vlan_number=vlan_number))

    def set_vlan_mpls_ip_state(self, vlan_number, state):
        self.put('/vlans/{vlan_number}/mpls-ip'.format(
            vlan_number=vlan_number
        ), data=state)

    def set_access_mode(self, interface_id):
        self.put("/interfaces/" + interface_id + '/port-mode', raw_data='access')

    def set_trunk_mode(self, interface_id):
        self.put("/interfaces/" + interface_id + '/port-mode', raw_data='trunk')

    def set_bond_access_mode(self, bond_number):
        self.put("/bonds/" + str(bond_number) + '/port-mode', raw_data='access')

    def set_bond_trunk_mode(self, bond_number):
        self.put("/bonds/" + str(bond_number) + '/port-mode', raw_data='trunk')

    def set_access_vlan(self, interface_id, vlan):
        self.put("/interfaces/" + interface_id + '/access-vlan', raw_data=str(vlan))

    def reset_interface(self, interface_id):
        self.put("/interfaces/" + interface_id)

    def unset_interface_access_vlan(self, interface_id):
        self.delete("/interfaces/" + interface_id + '/access-vlan')

    def set_interface_native_vlan(self, interface_id, vlan):
        self.put("/interfaces/" + interface_id + '/trunk-native-vlan', raw_data=str(vlan))

    def unset_interface_native_vlan(self, interface_id):
        self.delete("/interfaces/" + interface_id + '/trunk-native-vlan')

    def set_bond_native_vlan(self, bond_number, vlan):
        self.put("/bonds/" + str(bond_number) + '/trunk-native-vlan', raw_data=str(vlan))

    def unset_bond_native_vlan(self, bond_number):
        self.delete("/bonds/" + str(bond_number) + '/trunk-native-vlan')

    def add_trunk_vlan(self, interface_id, vlan):
        self.post("/interfaces/" + interface_id + '/trunk-vlans', raw_data=str(vlan))

    def remove_trunk_vlan(self, interface_id, vlan):
        self.delete("/interfaces/" + interface_id + '/trunk-vlans/' + str(vlan))

    def add_bond_trunk_vlan(self, bond_number, vlan):
        self.post("/bonds/" + str(bond_number) + '/trunk-vlans', raw_data=str(vlan))

    def remove_bond_trunk_vlan(self, bond_number, vlan):
        self.delete("/bonds/" + str(bond_number) + '/trunk-vlans/' + str(vlan))

    def set_interface_description(self, interface_id, description):
        self.put("/interfaces/" + interface_id + '/description', raw_data=description)

    def unset_interface_description(self, interface_id):
        self.delete("/interfaces/" + interface_id + '/description')

    def set_bond_description(self, bond_number, description):
        self.put("/bonds/" + str(bond_number) + '/description', raw_data=description)

    def unset_bond_description(self, bond_number):
        self.delete("/bonds/" + str(bond_number) + '/description')

    def set_interface_mtu(self, interface_id, size):
        self.put("/interfaces/" + interface_id + '/mtu', raw_data=str(size))

    def unset_interface_mtu(self, interface_id):
        self.delete("/interfaces/" + interface_id + '/mtu')

    def set_bond_mtu(self, bond_number, size):
        self.put("/bonds/" + str(bond_number) + '/mtu', raw_data=str(size))

    def unset_bond_mtu(self, bond_number):
        self.delete("/bonds/" + str(bond_number) + '/mtu')

    def edit_interface_spanning_tree(self, interface_id, edge=None):
        data = {}
        if edge is not None:
            data["edge"] = edge

        self.put("/interfaces/" + interface_id + '/spanning-tree', data=data)

    def set_interface_state(self, interface_id, state):
        self.put("/interfaces/" + interface_id + '/shutdown', raw_data='true' if state is OFF else 'false')

    def unset_interface_state(self, interface_id):
        self.delete("/interfaces/" + interface_id + '/shutdown')

    def set_interface_auto_negotiation_state(self, interface_id, state):
        self.put("/interfaces/" + interface_id + '/auto-negotiation', raw_data='true' if state is ON else 'false')

    def unset_interface_auto_negotiation_state(self, interface_id):
        self.delete("/interfaces/" + interface_id + '/auto-negotiation')

    def add_bond(self, number):
        self.post("/bonds", data={'number': number})

    def remove_bond(self, number):
        self.delete("/bonds/" + str(number))

    def add_interface_to_bond(self, interface, bond_number):
        self.put("/interfaces/" + interface + '/bond-master', raw_data=str(bond_number))

    def remove_interface_from_bond(self, interface):
        self.delete("/interfaces/" + interface + '/bond-master')

    def set_bond_link_speed(self, number, speed):
        self.put("/bonds/{0}/link-speed".format(number), raw_data=speed)

    def edit_bond_spanning_tree(self, number, edge=None):
        data = {}
        if edge is not None:
            data["edge"] = edge

        self.put("/bonds/{0}/spanning-tree".format(number), data=data)

    def add_vrrp_group(self, vlan_number, group_id, ips=None, priority=None, hello_interval=None, dead_interval=None,
                       track_id=None, track_decrement=None):
        self.post("/vlans/{}/vrrp-groups".format(vlan_number),
                  data=dict(id=group_id, ips=[str(ip) for ip in ips], priority=priority,
                            hello_interval=hello_interval, dead_interval=dead_interval, track_id=track_id,
                            track_decrement=track_decrement))

    def remove_vrrp_group(self, vlan_number, group_id):
        self.delete("/vlans/{}/vrrp-groups/{}".format(vlan_number, group_id))

    def add_vlan_varp_ip(self, vlan_number, ip_network):
        self.post('/vlans/{vlan_number}/varp-ips'.format(
            vlan_number=vlan_number,
        ), raw_data=str(ip_network))

    def remove_vlan_varp_ip(self, vlan_number, ip_network):
        self.delete('/vlans/{vlan_number}/varp-ips/{ip_network}'.format(
            vlan_number=vlan_number,
            ip_network=ip_network
        ))

    def add_dhcp_relay_server(self, vlan_number, ip_address):
        self.post("/vlans/{}/dhcp-relay-server".format(
            vlan_number), raw_data=str(ip_address))

    def remove_dhcp_relay_server(self, vlan_number, ip_address):
        self.delete("/vlans/{}/dhcp-relay-server/{}".format(
            vlan_number, ip_address))

    def set_interface_lldp_state(self, interface_id, enabled):
        self.put("/interfaces/{}/lldp".format(interface_id),
                 raw_data=_get_json_boolean(enabled))

    def set_vlan_arp_routing_state(self, vlan_number, state):
        self.put('/vlans/{}/arp-routing'.format(vlan_number),
                 raw_data='true' if state is ON else 'false')

    def set_vlan_icmp_redirects_state(self, vlan_number, state):
        self.put('/vlans/{}/icmp-redirects'.format(vlan_number),
                 raw_data=_get_json_boolean(state))

    def set_vlan_ntp_state(self, vlan_number, state):
        self.put('/vlans/{}/ntp'.format(vlan_number),
                 raw_data=_get_json_boolean(state))

    def set_vlan_unicast_rpf_mode(self, vlan_number, mode):
        self.put('/vlans/{}/unicast-rpf-mode'.format(vlan_number),
                 raw_data=str(mode))

    def unset_vlan_unicast_rpf_mode(self, vlan_number):
        self.delete('/vlans/{}/unicast-rpf-mode'.format(vlan_number))

    def get_versions(self):
        return self.get("/versions").json()

    def get(self, relative_url):
        return self._retry_on_unknown_session(
            lambda: self.validated(
                self.requests.get(**self.request(relative_url))))

    def post(self, relative_url, data=None, raw_data=None):
        return self._retry_on_unknown_session(
            lambda: self.validated(
                self.requests.post(
                    data=raw_or_json(raw_data, data),
                    **self.request(relative_url))))

    def put(self, relative_url, data=None, raw_data=None):
        return self._retry_on_unknown_session(
            lambda: self.validated(
                self.requests.put(
                    data=raw_or_json(raw_data, data),
                    **self.request(relative_url))))

    def delete(self, relative_url):
        return self._retry_on_unknown_session(
            lambda: self.validated(
                self.requests.delete(**self.request(relative_url))))

    def request(self, relative_url=''):
        headers = {
            'Netman-Model': self.switch_descriptor.model,
            'Netman-Username': self.switch_descriptor.username,
            'Netman-Password': self.switch_descriptor.password,
            'Netman-Port': str(self.switch_descriptor.port),
            'Netman-Max-Version': str(self.max_version),
            'Netman-Verbose-Errors': "yes"
        }

        if len(self._next_proxies) > 0:
            headers["Netman-Proxy-Server"] = ",".join(self._next_proxies)

        if self.session_id:
            url = "{netman_url}/switches-sessions/{session_id}{path}".format(
                netman_url=self._proxy,
                session_id=self.session_id,
                path=relative_url
            )
            headers['Netman-Session-Id'] = self.session_id
        else:
            url = "{netman_url}/switches/{switch}{path}".format(
                netman_url=self._proxy,
                switch=self.switch_descriptor.hostname,
                path=relative_url)
        self.logger.info("Querying " + url)
        return {
            "url": url,
            "headers": headers,
        }

    def validated(self, req):
        if req.status_code >= 400:
            try:
                error = req.json()
            except Exception as e:
                self.logger.exception(e)
                raise Exception("{0}: {1}".format(req.status_code, req.content))

            if "error-class" in error:
                if "error-module" in error:
                    try:
                        module = importlib.import_module(error["error-module"])
                        exception = getattr(module, error["error-class"])()
                        exception.args = (error["error"], )
                        exception.message = error["error"]
                    except:
                        exception = NetmanException('{error-module}.{error-class}: {error}'.format(**error))
                else:
                    exception = getattr(__builtin__, error["error-class"])(error["error"])
            else:
                exception = Exception(error["error"])

            raise exception
        return req

    def _retry_on_unknown_session(self, operation):
        try:
            return operation()
        except UnknownSession as e:
            self.logger.warning("Could not perform operation, {}...  "
                                "Requesting a new session".format(e))
            self._connect()
            return operation()


def _get_json_boolean(state):
    return {True: "true", False: "false"}[state]



# Copyright 2016 Internap.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from hamcrest import assert_that, is_

from netman.core.objects.exceptions import UnknownInterface
from netman.core.objects.interface_states import ON
from tests import has_message
from tests.adapters.compliance_test_case import ComplianceTestCase


class UnsetInterfaceAutoNegotiationStateTest(ComplianceTestCase):
    _dev_sample = "juniper_qfx_copper"

    def test_removes_the_state_completely(self):
        self.try_to.set_interface_auto_negotiation_state(self.test_port, state=ON)

        self.client.unset_interface_auto_negotiation_state(self.test_port)
        interface = self.client.get_interface(self.test_port)

        assert_that(interface.auto_negotiation, is_(None))

    def test_fails_if_the_interface_does_not_exist(self):
        with self.assertRaises(UnknownInterface) as expect:
            self.client.unset_interface_auto_negotiation_state('ge-0/0/128')

        assert_that(expect.exception, has_message("Unknown interface ge-0/0/128"))
